#include "MeshLoaders/MeshLoader.h"

std::vector<Node> &MeshLoader::GetNodes() {
  return nodes;
}

std::vector<FiniteElement> &MeshLoader::GetFiniteElements() {
  return finiteElements;
}

std::vector<BoundaryFiniteElement> &MeshLoader::GetBoundaryFiniteElements() {
  return boundaryFiniteElements;
}


std::vector<int> MeshLoader::GetFiniteElementsByNodesId(int id1, int id2, int id3) {
  std::vector<int> result;
  auto resultIter = finiteElements.begin();
  auto resultEndIter = finiteElements.end();
  while (resultIter != resultEndIter) {
    resultIter = std::find_if(resultIter, resultEndIter, [id1, id2, id3](const FiniteElement& currentElement) { //!!! Копирование currentElement на каждой итерации
      auto iterBegin = currentElement.nodesId.begin();
      auto iterEnd = currentElement.nodesId.end();
      return std::find(iterBegin, iterEnd, id1) != iterEnd &&
          std::find(iterBegin, iterEnd, id2) != iterEnd &&
          std::find(iterBegin, iterEnd, id3) != iterEnd;
    });

    if (resultIter != resultEndIter) {
      result.push_back(resultIter->id);
      ++resultIter;
    }
  }

  return result;
}

std::vector<int> MeshLoader::GetFiniteElementsByEdge(int firstVerticeId, int secondVerticeId) {
  std::vector<int> result;
  auto resultIter = finiteElements.begin();
  auto resultEndIter = finiteElements.end();
  while (resultIter != resultEndIter) {
    resultIter =
        std::find_if(resultIter, resultEndIter, [firstVerticeId, secondVerticeId](const FiniteElement& currentElement) { //!!! Копирование currentElement на каждой итерации
          auto iterBegin = currentElement.nodesId.begin();
          auto iterEnd = currentElement.nodesId.end();
          return std::find(iterBegin, iterEnd, secondVerticeId) != iterEnd &&
              std::find(iterBegin, iterEnd, firstVerticeId) != iterEnd;
        });

    if (resultIter != resultEndIter) {
      result.push_back(resultIter->id);
      ++resultIter;
    }
  }

  return result;
}

std::unordered_set<int> MeshLoader::GetVertexNodesBySurfaceId(int surfaceId) {
  std::unordered_set<int> result;
  auto currentBoundaryElement = boundaryFiniteElements.begin();
  while (currentBoundaryElement != boundaryFiniteElements.end()) {
    currentBoundaryElement = std::find_if(currentBoundaryElement,
                                          boundaryFiniteElements.end(),
                                          [surfaceId](const BoundaryFiniteElement &elem) {
                                            return elem.surfaceId == surfaceId;
                                          });
    if (currentBoundaryElement != boundaryFiniteElements.end()) {
      for (const auto &elem: currentBoundaryElement->nodesId) {
        Node nodeById = nodes[elem];
        result.insert(nodeById.id);
      }
      ++currentBoundaryElement;
    }
  }
  return result;

}
std::vector<int> MeshLoader::GetFiniteElementsByMaterialId(int materialId) {
  std::vector<int> result;
  std::for_each(finiteElements.begin(), finiteElements.end(), [&materialId, &result](const FiniteElement &elem) {
    if (elem.materialId == materialId) {
      result.push_back(elem.id);
    }
  });
  return result;
}

std::vector<int> MeshLoader::GetBboundaryFiniteElementsBySurfaceId(int id) {
  std::vector<int> result;
  std::for_each(boundaryFiniteElements.begin(),
                boundaryFiniteElements.end(),
                [&id, &result](const BoundaryFiniteElement &elem) {
                  if (elem.surfaceId == id) {
                    result.push_back(elem.id);
                  }
                });
  return result;
}
std::vector<std::unordered_set<int>> MeshLoader::GetNodesNeighbors() {
  std::vector<std::unordered_set<int>> neighbors(nodes.size() + 1);
  for (const auto &elem: boundaryFiniteElements) {
    for (auto nodeId: elem.nodesId) {
      for (auto anthNodeId: elem.nodesId) {
        if (nodeId != anthNodeId) {
          neighbors[nodeId].insert(anthNodeId);
        }
      }
    }
  }

  return neighbors;
}

float MeshLoader::minAngle(const BoundaryFiniteElement &element) {
  float minAngle = -1;
  for (auto cur: element.nodesId) {
    auto coordNodes = element.nodesId
        | std::ranges::views::filter([cur](auto currentId) { return cur != currentId; })
        | std::ranges::views::transform([this](int currentId) { return nodes[currentId - 1].coordinates; });

    std::vector<std::vector<float>> coordinatesOfNodes;
    std::ranges::copy(coordNodes, std::back_inserter(coordinatesOfNodes));
    float currentAngle = angle(nodes[cur - 1].coordinates, coordinatesOfNodes);
    if (currentAngle < minAngle || minAngle == -1) {
      minAngle = currentAngle;
    }
  }

  return minAngle;
}

void MeshLoader::SortByMinAngle() {
  auto res = std::ranges::sort(boundaryFiniteElements,
                               std::ranges::less{},
                               [this](BoundaryFiniteElement &element) { return minAngle(element); });

}
void MeshLoader::InsertMiddle() {
  auto func = [](const Edge &elem) {
    
    //!!! 1. Плохая хеш-функция. Нужно использовать hashCombain из лекции.
    //!!! 2. Хеш-функция должна считаться для отсортированных данных. Иначе для ребер (1,2) и (2,1) получаются разные хеш-значения, хотя это одно ребро.
    
    return std::hash<int>()(elem.edgeNodesId.first) ^ std::hash<int>()(elem.edgeNodesId.second);
  };
  std::unordered_set<Edge, decltype(func)> edges;
  for (auto &element: finiteElements) {
    for (int first = 0; first < 4; ++first) {
      for (auto second = first + 1; second < 4; ++second) {
        Edge currentEdge(element.nodesId[first], element.nodesId[second]);
        if (edges.insert(currentEdge).second != 0) {
          Node newNode = createMiddleNode(currentEdge);
          currentEdge.middleNodeId = newNode.id;
          nodes.push_back(newNode);
          element.nodesId.push_back(newNode.id);
        } else {
          element.nodesId.push_back(currentEdge.middleNodeId);
        }
      }
    }
  }
  for (auto &element: boundaryFiniteElements) {
    for (auto first = 0; first < 3; ++first) {
      for (auto second = first + 1; second < 3; ++second) {
        auto edgeIter = edges.find({element.nodesId[first], element.nodesId[second]});
        if (edgeIter == edges.end()) {
          edgeIter = edges.find({element.nodesId[second], element.nodesId[first]});
        }
        element.nodesId.push_back(edgeIter->middleNodeId);
      }
    }
  }
}
Node MeshLoader::createMiddleNode(const Edge &edge) {
  std::vector<float> coord(3);
  coord[0] = (nodes[edge.edgeNodesId.first - 1].coordinates[0] +
      nodes[edge.edgeNodesId.second - 1].coordinates[0]) / 2;

  coord[1] = (nodes[edge.edgeNodesId.first - 1].coordinates[1] +
      nodes[edge.edgeNodesId.second - 1].coordinates[1]) / 2;

  coord[2] = (nodes[edge.edgeNodesId.first - 1].coordinates[2] +
      nodes[edge.edgeNodesId.second - 1].coordinates[2]) / 2;

  Node newNode{nodes.size() + 1, coord, false};
  return newNode;
}

float scalarMultiply(const std::vector<float> &p1, const std::vector<float> &p2) {
  float result = 0;
  for (int i = 0; i < p1.size(); ++i) {
    result += p1[i] * p2[i];
  }

  return result;
}

float distance(const std::vector<float> &p1, const std::vector<float> &p2) {
  float result = 0;
  for (int i = 0; i < p1.size(); ++i) {
    result += (p1[i] - p2[i] * (p1[i] - p2[i]));
  }

  return std::sqrt(result);
}

float angle(const std::vector<float> &p, const std::vector<std::vector<float>> &otherPoints) {
  auto x = otherPoints[0];
  auto y = otherPoints[1];

  return ((x[0] - p[0]) * (y[0] - p[0]) + (x[1] - p[1]) * (y[1] - p[1]) + (x[2] - p[2]) * (y[2] - p[2]))
      / (distance(x, p) * (distance(y, p)));
}